import unittest
from unittest.mock import patch
from mistletoe import span_token


class TestBranchToken(unittest.TestCase):
    def setUp(self):
        self.addCleanup(
            lambda: span_token._token_types.value.__setitem__(-1, span_token.RawText)
        )
        patcher = patch("mistletoe.span_token.RawText")
        self.mock = patcher.start()
        span_token._token_types.value[-1] = self.mock
        self.addCleanup(patcher.stop)

    def _test_parse(self, token_cls, raw, arg, **kwargs):
        token = next(iter(span_token.tokenize_inner(raw)))
        self.assertIsInstance(token, token_cls)
        self._test_token(token, arg, **kwargs)

    def _test_token(self, token, arg, children=True, **kwargs):
        for attr, value in kwargs.items():
            self.assertEqual(getattr(token, attr), value)
        if children:
            self.mock.assert_any_call(arg)


class TestStrong(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.Strong, "**some text**", "some text")
        self._test_parse(span_token.Strong, "__some text__", "some text")


class TestEmphasis(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.Emphasis, "*some text*", "some text")
        self._test_parse(span_token.Emphasis, "_some text_", "some text")


class TestInlineCode(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.InlineCode, "`some text`", "some text")


class TestStrikethrough(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.Strikethrough, "~~some text~~", "some text")


class TestLink(TestBranchToken):
    def test_parse(self):
        self._test_parse(
            span_token.Link, "[name 1](target1)", "name 1", target="target1", title=""
        )

    def test_parse_multi_links(self):
        tokens = iter(span_token.tokenize_inner("[n1](t1) & [n2](t2)"))
        self._test_token(next(tokens), "n1", target="t1")
        self._test_token(next(tokens), " & ", children=False)
        self._test_token(next(tokens), "n2", target="t2")

    def test_parse_children(self):
        token = next(iter(span_token.tokenize_inner("[![alt](src)](target)")))
        child = next(iter(token.children))
        self._test_token(child, "alt", src="src")


class TestAutoLink(TestBranchToken):
    def test_parse(self):
        self._test_parse(
            span_token.AutoLink,
            "<ftp://foo.com>",
            "ftp://foo.com",
            target="ftp://foo.com",
        )


class TestImage(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.Image, "![alt](link)", "alt", src="link")
        self._test_parse(
            span_token.Image, '![alt](link "title")', "alt", src="link", title="title"
        )

    def test_no_alternative_text(self):
        self._test_parse(span_token.Image, "![](link)", "", children=False, src="link")


class TestEscapeSequence(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.EscapeSequence, "\\*", "*")

    def test_parse_in_text(self):
        tokens = iter(span_token.tokenize_inner("some \\*text*"))
        self._test_token(next(tokens), "some ", children=False)
        self._test_token(next(tokens), "*")
        self._test_token(next(tokens), "text*", children=False)


class TestRawText(unittest.TestCase):
    def test_attribute(self):
        token = span_token.RawText("some text")
        self.assertEqual(token.content, "some text")

    def test_no_children(self):
        token = span_token.RawText("some text")
        with self.assertRaises(AttributeError):
            token.children


class TestLineBreak(unittest.TestCase):
    def test_parse(self):
        (token,) = span_token.tokenize_inner("  \n")
        self.assertIsInstance(token, span_token.LineBreak)


class TestContains(unittest.TestCase):
    def test_contains(self):
        token = next(iter(span_token.tokenize_inner("**with some *emphasis* text**")))
        self.assertTrue("text" in token)
        self.assertTrue("emphasis" in token)
        self.assertFalse("foo" in token)



"""
Make mistletoe easier to import.
"""

__version__ = "0.8.2"
__all__ = [
    "html_renderer",
    "ast_renderer",
    "block_token",
    "block_tokenizer",
    "span_token",
    "span_tokenizer",
]

from mistletoe.block_token import Document
from mistletoe.base_renderer import BaseRenderer  # noqa: F401
from mistletoe.html_renderer import HTMLRenderer


def markdown(iterable, renderer=HTMLRenderer):
    """
    Output HTML with default settings.
    Enables inline and block-level HTML tags.
    """
    with renderer() as renderer:
        return renderer.render(Document(iterable))



from unittest import TestCase
from mistletoe.core_tokens import (
    MatchObj,
    Delimiter,
    follows,
    shift_whitespace,
    is_control_char,
    deactivate_delimiters,
    preceded_by,
    succeeded_by,
)


class TestCoreTokens(TestCase):
    def test_match_obj(self):
        match = MatchObj(0, 2, (0, 1, "a"), (1, 2, "b"))
        self.assertEqual(match.start(), 0)
        self.assertEqual(match.start(1), 0)
        self.assertEqual(match.start(2), 1)
        self.assertEqual(match.end(), 2)
        self.assertEqual(match.end(1), 1)
        self.assertEqual(match.end(2), 2)
        self.assertEqual(match.group(), "ab")
        self.assertEqual(match.group(1), "a")
        self.assertEqual(match.group(2), "b")

    def test_delimiter(self):
        delimiter = Delimiter(4, 6, "abcd**")
        self.assertEqual(delimiter.type, "**")
        self.assertEqual(delimiter.number, 2)
        self.assertEqual(delimiter.active, True)
        self.assertEqual(delimiter.start, 4)
        self.assertEqual(delimiter.end, 6)

    def test_delimiter_remove_left(self):
        delimiter = Delimiter(4, 6, "abcd**")
        self.assertTrue(delimiter.remove(1, left=True))
        self.assertEqual(delimiter.number, 1)
        self.assertEqual(delimiter.start, 5)
        self.assertEqual(delimiter.end, 6)

    def test_delimiter_remove_right(self):
        delimiter = Delimiter(4, 6, "abcd**")
        self.assertTrue(delimiter.remove(1, left=False))
        self.assertEqual(delimiter.number, 1)
        self.assertEqual(delimiter.start, 4)
        self.assertEqual(delimiter.end, 5)

    def test_delimiter_remove_empty(self):
        delimiter = Delimiter(4, 6, "abcd**")
        self.assertFalse(delimiter.remove(2))

    def test_follows(self):
        string = "(foobar)"
        self.assertTrue(follows(string, 6, ")"))
        self.assertFalse(follows(string, 6, "("))
        self.assertFalse(follows(string, 7, ")"))

    def test_shift_whitespace(self):
        string = " \n\t\rfoo"
        self.assertEqual(shift_whitespace(string, 0), 4)
        self.assertEqual(shift_whitespace("", 0), 0)

    def test_is_control_char(self):
        char = chr(0)
        self.assertTrue(is_control_char(char))
        self.assertFalse(is_control_char("a"))

    def test_deactivate_delimiters(self):
        s = "abc"
        delimiters = [Delimiter(0, 1, s), Delimiter(1, 2, s), Delimiter(2, 3, s)]
        deactivate_delimiters(delimiters, 2, "b")
        self.assertTrue(delimiters[0].active)
        self.assertFalse(delimiters[1].active)
        self.assertTrue(delimiters[2].active)

    def test_preceded_by(self):
        whitespace = " \t\n\r"
        self.assertTrue(preceded_by(1, " abc", whitespace))
        self.assertTrue(preceded_by(0, "aabc", whitespace))
        self.assertFalse(preceded_by(1, "aabc", whitespace))
        self.assertFalse(preceded_by(0, "aabc", "abc"))

    def test_succeeded_by(self):
        whitespace = " \t\n\r"
        self.assertTrue(succeeded_by(3, "abc ", whitespace))
        self.assertTrue(succeeded_by(4, "abcc", whitespace))
        self.assertFalse(succeeded_by(3, "abcc", whitespace))
        self.assertFalse(succeeded_by(4, "abcc", "abc"))
